/* determine the ancestral orientations */

#include "util.h"
#include "species.h"

#define PO  0x1
#define IV  0x2
#define UN  0x4

struct tnode {
	int nid, state;
	char name[20];
	double blen;
	struct tnode *father, *lchild, *rchild;
};

struct result_list {
	int lev;
	char spechr[100];
	int beg, end;
	char *branch;
	int rm, rb, re;
	char refchr[100];
	struct result_list *next;
};

static struct result_list **reslist;
static struct tnode **leaf;
static struct tnode *Ancestor = NULL;
static int *Beg;
static int *End;
static char **Chrom;

int ori2int(char x) {
	int o;
	switch (x) {
		case '+': o = PO;	break;
		case '-': o = IV;	break;
		case '?': o = UN;	break;
		default: fatalf("illegal ori: %c", x);
	}
	return o;
}

void remove_tree(struct tnode *node) { 
	if (node == NULL)
		return ;
	remove_tree(node->lchild); 
	remove_tree(node->rchild); 
	free(node);
}

struct tnode *get_tree2(char *str) {
	struct tnode *stack[200], *p, *q;
	char ch, buf[20], buf2[20], *endp;
	double dc;
	int k, top, bidx, b2idx, count, nidcount, j;
	
	p = NULL;
	k = top = bidx = b2idx = nidcount = j = 0;
	count = 1;
	ch = str[k];
	
	while (ch != '\0' && ch != ';') {
		if (isalpha(ch)) {
			p = (struct tnode *)ckalloc(sizeof(struct tnode));
			p->blen = 0;
			p->nid = nidcount++;
			p->lchild = NULL;
			p->rchild = NULL;
			buf[bidx++] = ch;
			ch = str[++k];
			while (ch != ',' && ch != ')') {
				buf[bidx++] = ch;
				ch = str[++k];
			}
			buf[bidx] = '\0';
			bidx = 0;
			strcpy(p->name, buf);
			for (j = 0; j < Spesz; j++)
				if (same_string(Spename[j], p->name))
					break;
			leaf[j] = p;
		}
		if (ch == ':') {
			ch = str[++k];
			while (ch != ',' && ch != ')') {
				buf2[b2idx++] = ch;
				ch = str[++k];
			}
			buf2[b2idx] = '\0';
			dc = strtod(buf2, &endp);
			b2idx = 0;
		} 
    switch (ch) {
			case '(': {
				p = (struct tnode *)ckalloc(sizeof(struct tnode));
				sprintf(buf2, "IN%d", count++);
				strcpy(p->name, buf2);
				p->blen = 0;
				p->father = NULL;
				p->lchild = NULL;
				p->rchild = NULL;
				p->nid = nidcount++;
				stack[top++] = p;
				break;
			}
			case ',': {
				q = stack[top - 1];
				q->lchild = p;
				p->father = q;
				p->blen = dc;
				break;
			}
			case ')': {
				q = stack[--top];
				q->rchild = p;
				p->father = q;
				p = q;
				if (str[k+1] == '@') {
					Ancestor = q;
					++k;
					break;
				}
				/* get internal node name */
				bidx = 0;
				while (isalpha(str[k+1])) {
					++k;
					ch = str[k];
					buf[bidx++] = ch;
				}
				if (bidx != 0) {
					buf[bidx] = '\0';
					bidx = 0;
					strcpy(q->name, buf);
				}
				break;
			}
			default: ; /* do nothing */
		}
		ch = str[++k];
	}
	return p;
}

void set_code(struct tnode *node, char* name, int code) {
	if (node == NULL)
		return ;
	if (same_string(node->name, name)) {
		node->state = code;
		return ;
	}
	set_code(node->lchild, name, code);
	set_code(node->rchild, name, code);
}

void init_code(struct tnode *node) {
	if (node == NULL)
		return ;
	node->state = 0;
	init_code(node->lchild);
	init_code(node->rchild);
}

int set_ances(struct tnode *node) {
	int l, r, lar;
	
	if (node->lchild == NULL || node->rchild == NULL)
		return node->state;
	l = set_ances(node->lchild);
	r = set_ances(node->rchild);
	lar = l & r;
	if ( lar == PO || lar == IV || lar == (PO|UN) || lar == (IV|UN)) {
		node->state = (l & r);
		return node->state;
	}
	node->state = (l | r);
	return node->state;
}

void update_with_outgroup(struct tnode *node, struct tnode *outgrp) {
	int a, b, c;
	a = node->state;
	b = outgrp->state;
	c = a & b;
	if (c == PO || c == IV)
		node->state = c;
}

void apply_outgroup(struct tnode *node) {
	struct tnode *cur, *outgrp;
	cur = node;
	while (cur->father != NULL) {
		if (cur->father->lchild == cur)
			outgrp = cur->father->rchild;
		else
			outgrp = cur->father->lchild;
		update_with_outgroup(node, outgrp);
		cur = cur->father;
	}
}


void modify_code(struct tnode *dad, struct tnode *son) {
	int d, s, das;
	d = dad->state;
	s = son->state;
	das = d & s;
	if (das == PO || das == IV) 
		son->state = das;
	else if (s == (PO|UN) || s == (IV|UN)) //child orientation: +x or -x
			son->state = d;
}

void change_internal_code(struct tnode *node) {
	if (node->lchild == NULL || node->rchild == NULL)
		return;
	modify_code(node, node->lchild);
	modify_code(node, node->rchild);
	change_internal_code(node->lchild);
	change_internal_code(node->rchild);
}

void add_to_list(int speidx, int lev, char *sc, 
								int b, int e, char *brch,
								char *rc, int rb, int re) {
	struct result_list *p, *q;
	q = (struct result_list *)ckalloc(sizeof(struct result_list));
	q->next = NULL;
	strcpy(q->spechr, sc);
	q->lev = lev;
	q->beg = b;
	q->end = e;
	q->rm = 0;
	q->branch = brch;
	strcpy(q->refchr, rc);
	q->rb = rb;
	q->re = re;
	
	if (reslist[speidx] == NULL)
		reslist[speidx] = q;
	else {
		for (p = reslist[speidx]; p->next != NULL; p = p->next)
			; /* do nothing */
		p->next = q;
	}
}

void free_result_list(int speidx) {
	struct result_list *p, *q;
	p = reslist[speidx];
	while (p != NULL) {
		q = p->next;
		free(p->branch);
		free(p);
		p = q;
	}
}

static void postprocess_list(int speidx) {
	struct result_list *p, *q;
	for (p = reslist[speidx]; p != NULL; p = p->next) {
		if (p->rm == 1)
			continue;
		for (q = reslist[speidx]; q != NULL; q = q->next) {
			if (q != p && same_string(p->spechr, q->spechr) && q->rm == 0
						&& same_string(p->refchr, q->refchr)) {
				if (p->beg >= q->beg && p->end <= q->end)
					p->rm = 1;
				else if (p->beg <= q->beg && p->end >= q->end)
					q->rm = 1;
				else if ((p->beg < q->beg && p->end > q->beg)
						|| (p->beg < q->end && p->end > q->end) ) {
						if (p->end - p->end > q->end - q->beg)
							q->rm = 1;
						else
							p->rm = 1;
				}
			}
		}
	}
}

void print_result_list(int speidx) {
	char buf[500];
	FILE *fp;
	struct result_list *p;
	
	sprintf(buf, "%s.micro.inv", Spename[speidx]);
	fp = ckopen(buf, "w");
	for (p = reslist[speidx]; p != NULL; p = p->next)
		if (p->rm == 0)
			fprintf(fp, "%s.%s:%d-%d %s\t[%s.%s:%d-%d]\n", 
						Spename[speidx], p->spechr, p->beg, p->end, p->branch,
						Spename[0], p->refchr, p->rb, p->re);
	fclose(fp);
}

char *assign_to_branch(struct tnode *t) {
	struct tnode *p;
	for (p = t; p != NULL && p->father != NULL; p = p->father) {
		if (p->state != p->father->state)
			break;
	}
	return copy_string(p->name);
}

void output_inversion(int anc) {
	int j, rs, lev = 1; //hard-coded to be 1
	char *branch;
	struct tnode *t;
	
	rs = ref_spe_idx();
	
	for (j = 0; j < Spesz; j++) {
		t = leaf[j];
		if (t->state != UN) {
			if (t->state != anc) {
				branch = assign_to_branch(t);
				add_to_list(j, lev, Chrom[j], Beg[j], End[j], branch, Chrom[rs], Beg[rs], End[rs]);
			}
		}
	}
}

void *ckcalloc(size_t nelem, size_t elsize) {
	void *space;
	if ((space = calloc(nelem, elsize)) == NULL) {
		fprintf(stderr, "calloc() failed: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}
	return(space);
}

static void allocate_arrays(void) {
	int i;
	/* one dimensional arrays */
	reslist = (struct result_list **) ckcalloc(Spesz, sizeof(struct result_list *));
	leaf = (struct tnode **) ckcalloc(Spesz, sizeof(struct tnode *));
	Beg = (int *) ckcalloc(Spesz, sizeof(int));
	End = (int *) ckcalloc(Spesz, sizeof(int));
	/* two dimensional array */
	Chrom = (char **) ckcalloc(Spesz, sizeof(char *));
	for (i = 0; i < Spesz; i++) {
		Chrom[i] = (char *) ckcalloc(20, sizeof(char));
	}
}

static void free_arrays(void) {
	int i;
	/* one dimensional arrays */
	free(reslist);
	free(leaf);
	free(Beg);
	free(End);
	/* two dimensional array */
	for (i = 0; i < Spesz; i++) {
		free(Chrom[i]);
	}
	free(Chrom);
}

int main (int argc, char* argv[]) {
	FILE *ifp;
	char buf[500], spe[20], chr[20];
	char orient;
	char *ref;
	struct tnode *root;
	int code, i=0, j, a, b, e, rs;

	if (argc != 3)
  	fatal("arg = config-file inversion-list");

	get_spename(argv[1]);
	get_treestr2(argv[1]);
	rs = ref_spe_idx();
	ref = Spename[rs];

	allocate_arrays();
	root = get_tree2(Treestr2);
	ifp = ckopen(argv[2], "r");

  while (fgets(buf, 500, ifp)) {
		if (buf[0] == '\n' || buf[0] == '>')
			continue;
        if (buf[0] == '#') {
            init_code(root);
            i = 0;
            continue;
        }
		if (buf[0] == '\n' || buf[0] == '-')
			continue;
		if (sscanf(buf, "ANCESTOR: %c", &orient) == 1) {
			a = ori2int(orient);
			Ancestor->state = a;
			change_internal_code(Ancestor);
			output_inversion(a);
			continue;
		}
		i++;
		if (strchr(buf, '?') != NULL) {
			if (sscanf(buf, " %c %s", &orient, spe) != 2)
				fatalf("cannot parse %s", buf);
			code = ori2int(orient);
			set_code(root, spe, code);
		}
		else if (sscanf(buf, " %c %[^.].%[^:]:%d-%d %*s", &orient, spe, chr, &b, &e) != 5)
			fatalf("cannot parse %s", buf);
		else {
			code = ori2int(orient);
			set_code(root, spe, code);
			j = spe_idx(spe);
			strcpy(Chrom[j], chr);
			Beg[j] = b;
			End[j] = e;
		}
		if (i == Spesz) {
			set_ances(root);
			apply_outgroup(Ancestor);
		}
	}
	fclose(ifp);
    
	for (j = 0; j < Spesz; j++) {
		postprocess_list(j);
		if (Spetag[j] != 2)
			print_result_list(j);
		free_result_list(j);
	}
	remove_tree(root);
	free_arrays();
  
	return 0;
}
