#include "util.h"
#include "species.h"

#define PO  0x1
#define IV  0x2
#define UN  0x4

struct tnode {
	int nid, state;
	char name[20];
	double blen;
	struct tnode *father, *lchild, *rchild;
};

struct tnode *Ancestor = NULL;

void int2ori(char *s, int x) {
	switch (x) {
		case PO:				strcpy(s, "+");		break;
		case IV:				strcpy(s, "-");		break;
		case PO|IV:			strcpy(s, "+-");	break;
		case UN:				strcpy(s, "?");		break;
		case PO|UN:			strcpy(s, "+?");	break;
		case IV|UN:			strcpy(s, "-?");	break;
		case PO|IV|UN:  strcpy(s, "+-?");	break;
		default: fatalf("illegal val: %d", x);
	}
}

int ori2int(char x) {
	int o;
	switch (x) {
		case '+': o = PO;	break;
		case '-': o = IV;	break;
		case '?': o = UN;	break;
		default: fatalf("illegal ori: %c", x);
	}
	return o;
}

void remove_tree(struct tnode *node) { 
	if (node == NULL)
		return ;
	remove_tree(node->lchild); 
	remove_tree(node->rchild); 
	free(node);
}

struct tnode *get_tree(char *str) { 
	struct tnode *stack[200], *p, *q; 
	char ch, buf[20], buf2[20], *endp; 
	double dc;
	int k, top, bidx, b2idx, count, nidcount; 

	p = NULL;
	k = top = bidx = b2idx = nidcount = 0;
	count = 1;
	
	ch = str[k]; 
	while (ch != '\0' && ch != ';') {
		if (isalpha(ch)) { 
			p = (struct tnode *)ckalloc(sizeof(struct tnode)); 
			p->blen = 0; 
			p->nid = nidcount++; 
			p->lchild = p->rchild = NULL; 
			buf[bidx++] = ch; 
			ch = str[++k]; 
			while (ch != ',' && ch != ')') { 
				buf[bidx++] = ch; 
				ch = str[++k]; 
			} 
			buf[bidx] = '\0'; 
			bidx = 0; 
			strcpy(p->name, buf); 
		} 
		if (ch == ':') { 
			ch = str[++k]; 
			while (ch != ',' && ch != ')') { 
				buf2[b2idx++] = ch; 
				ch = str[++k]; 
			} 
			buf2[b2idx] = '\0'; 
			dc = strtod(buf2, &endp); 
			b2idx = 0; 
		} 
		switch (ch) { 
			case '(': { 
				p = (struct tnode *)ckalloc(sizeof(struct tnode)); 
				sprintf(buf2, "IN%d", count++); 
				strcpy(p->name, buf2); 
				p->blen = 0; 
				p->father = p->lchild = p->rchild = NULL; 
				p->nid = nidcount++; 
				stack[top++] = p; 
				break; 
			} 
			case ',': { 
				q = stack[top - 1]; 
				q->lchild = p; 
				p->father = q; 
				p->blen = dc; 
				break; 
			} 
			case ')': { 
				q = stack[--top];
				q->rchild = p;
				p->father = q;
				p = q;
				if (str[k+1] == '@') {
					Ancestor = q;
					++k;
					break;
				}
				/* get internal node name */
				bidx = 0;
				while (isalpha(str[k+1])) {
					++k;
					ch = str[k];
					buf[bidx++] = ch;
				}
				if (bidx != 0) {
					buf[bidx] = '\0';
					bidx = 0;
					strcpy(q->name, buf);
				}
				break;
			 } 
			default: ; /* do nothing */ 
		} 
		ch = str[++k]; 
	} 
	return p;
}

void set_code(struct tnode *node, char *name, int code) {
	if (node == NULL)
  	return ;
	if (same_string(node->name, name)) {
		node->state = code;
			return ;
	}
	set_code(node->lchild, name, code);
	set_code(node->rchild, name, code);
}

void init_code(struct tnode *node) {
	if (node == NULL)
		return ;
	node->state = 0;
	init_code(node->lchild);
	init_code(node->rchild);
}

int set_ances(struct tnode *node) {
	int l, r, lar;
	if (node->lchild == NULL || node->rchild == NULL)
		return node->state;
	l = set_ances(node->lchild);
	r = set_ances(node->rchild);
	lar = l & r;
	if (lar == PO || lar == IV || lar == (PO|UN) || lar == (IV|UN)) {
		node->state = (l & r);
		return node->state;
	}
	node->state = (l | r);
	return node->state;
}

void update_with_outgroup(struct tnode *node, struct tnode *outgrp) {
	int a, b, c;
	a = node->state;
	b = outgrp->state;
	c = a & b;
	if (c == PO || c == IV)
		node->state = c;
}

void apply_out_group(struct tnode *node) {
	struct tnode *cur, *outgrp;
	cur = node;
	while (cur->father != NULL) {
		if (cur->father->lchild == cur)
			outgrp = cur->father->rchild;
		else
			outgrp = cur->father->lchild;
		update_with_outgroup(node, outgrp);
		cur = cur->father;
	}
}

int main (int argc, char* argv[]) {
	FILE *ifp;
	char buf[500], spe[20], fun[5], fun2[5];
	char orient;
	struct tnode *root;
	int code, i, anc, hcnt;

	i = hcnt = 0;

	if (argc != 3)
  	fatal("arg = config-file inversion-list");

	get_spename(argv[1]);
	get_treestr2(argv[1]);
	root = get_tree(Treestr2);
	ifp = ckopen(argv[2], "r");

	while (fgets(buf, 500, ifp)) {
		if (buf[0] == '#' || buf[0] == '>' || buf[0] == '\n') {
			printf("%s", buf);
			init_code(root);
			i = 0;
			continue;
		}
		i++;
		printf("%s", buf);
		sscanf(buf, "  %c %[^.].%*s %*s", &orient, spe);
		code = ori2int(orient);
		set_code(root, spe, code);
		if (i == Spesz) {
			anc = set_ances(root);
			apply_out_group(Ancestor);
			anc = Ancestor->state;
			int2ori(fun, anc);
			printf("----------------\n");
			if (anc == PO || anc == IV) {
				printf("ANCESTOR: %s ", fun);
				if (anc == IV)
					hcnt++;
			}
			if (anc == (PO | UN) || anc == (IV | UN)) {
				anc = anc - UN;
				int2ori(fun2, anc);
				printf("ANCESTOR: %s ", fun2);
			}
			if (anc == (PO | IV) || anc == (PO | IV | UN)) {
				anc = PO;
				int2ori(fun2, anc);
				printf("ANCESTOR: %s ", fun2);
			}
			printf("\n\n");
		}
	}
	
	fclose(ifp);
  remove_tree(root);
	
	return 0; 
}
