
#include "util.h"
#include "species.h"
#include "base2.h"

#define BIGNUM 	300000000
#define WIN			200000
#define AAFEW 	0.1

struct net_list {
	char type[20];
	int fbeg, fend;
	int sbeg, send;
	struct net_list *next;
};

int rs;

static struct block_list *get_inv_block(struct block_list *block, int idx, 
										char *chr, int b, int e, int left, int right, 
										struct block_list **lst) {
	struct block_list *p, *fst;
	struct seg_list *ps;

	fst = *lst = NULL;
	
	for (p = block; p != NULL; p = p->next) {
		ps = p->speseg[idx];
		if (same_string(chr, ps->chr) && 
			((b <= ps->beg && e >= ps->end) || (b >= ps->beg && e <= ps->end)
			 || (b < ps->beg && e > ps->beg) || (b < ps->end && e > ps->end ))) {
			if ((b < p->left && (p->left - b) > AAFEW * (e - b)) 
				|| ((e > p->right) && (e - p->right) > AAFEW * (e - b)) )
				continue;
			if ((left > ps->beg && (left - ps->beg) > AAFEW * (ps->end - ps->beg))
				|| (right < ps->end && (ps->end - right) > AAFEW * (ps->end - ps->beg) ))
				continue;
			if (fst == NULL)
				fst = *lst = p;
			else
				*lst = p;
		}
	}
	return fst;
}

static struct block_list *find_pos(struct block_list *block, int idx, int b, int id) {
	struct block_list *p;
	struct seg_list *ps, *pt;
	
	if (id < block->id || (id == block->id && b <= block->speseg[idx]->beg))
		return NULL;
	for (p = block; p != NULL; p = p->next) {
		ps = p->speseg[idx];
		if (p->next == NULL)
			break;
		pt = p->next->speseg[idx];
		if (id == p->id && b >= ps->beg) {
			if ((id == p->next->id && b < pt->beg) || (id < p->next->id))
				break;
		}
		else if ((id >= p->id && id < p->next->id) 
					|| (id > p->id && id == p->next->id && b < pt->beg))
				break;
	}
	return p;
}

static void add_info(struct block_list *block, char *species, char *chr,
						int b, int e, int rs, int left, int right) {
	struct seg_list *p;
	int idx, bb, ee;
	
	idx = spe_idx(species);
	p = (struct seg_list *)ckalloc(sizeof(struct seg_list));
	p->cidlist = NULL;
	p->next = NULL;
	strcpy(p->chr, chr);
	p->beg = b;
	p->end = e;
	if (idx != rs) {
		if (Spetag[idx] != 2) {
			block->left = MAX(block->left, left);
			block->right = MIN(block->right, right);
		}
		p->orient = '-';
	}
	else 
		p->orient = '+';
	if (block->speseg[idx] == NULL) 
		block->speseg[idx] = p;
	else {
		if (!same_string(block->speseg[idx]->chr, chr))
			fatalf("cannot happen: %s.%s:%d-%d", species, chr, b, e);
		bb = block->speseg[idx]->beg;
		ee = block->speseg[idx]->end;
		if (idx != rs) {
			if (bb - e > WIN || b - ee > WIN) {
				if (e - b > ee - bb) {
					block->speseg[idx]->beg = b;
					block->speseg[idx]->end = e;
				}
			}
			else {
				block->speseg[idx]->beg = MIN(block->speseg[idx]->beg, b);
				block->speseg[idx]->end = MAX(block->speseg[idx]->end, e);
			}
		}
		else {
				block->speseg[idx]->beg = MIN(block->speseg[idx]->beg, b);
				block->speseg[idx]->end = MAX(block->speseg[idx]->end, e);
		}
		free(p);
	}
}

static void merge_blocks(struct block_list *from, struct block_list *to) {
	struct block_list *p, *q, *terminal;
	int i;
	p = from;
	terminal = to->next;
	while(p->next != terminal) {
		q = p->next;
		p->next = q->next;
		q->next = NULL;
		p->left = MIN(p->left, q->left);
		p->right = MAX(p->right, q->right);
		for (i = 0; i < Spesz; i++) {
			if (p->speseg[i] == NULL)
				p->speseg[i] = q->speseg[i];
			else if (q->speseg[i] == NULL)
					continue;
			else { 
				if (!same_string(p->speseg[i]->chr, q->speseg[i]->chr)) 
					continue; // just leave one
				p->speseg[i]->beg = MIN(p->speseg[i]->beg, q->speseg[i]->beg);
				p->speseg[i]->end = MAX(p->speseg[i]->end, q->speseg[i]->end);
			}
		}
	}
}

static struct block_list *find_original_block(struct block_list *block, int id) {
	struct block_list *blk;
	
	for (blk = block; blk != NULL; blk = blk->next)
		if (blk->id == id)
			break;
	if (blk == NULL)
		fatalf("id %d not exist", id);
	
	return blk;
}

static void check_covered(struct block_list *invlist, int i, int j, struct block_list *block) {
	struct block_list *blk;
	struct seg_list *ref, *sec, *newsg;
	int b, e, tmp, tag1, tag2, len, c, f, k;
	
	tag1 = tag2 = 0;
	fprintf(stderr, "=> %s %s %d\n", Spename[i], Spename[j], invlist->id);
	blk = find_original_block(block, invlist->id);
	
	ref = invlist->speseg[i];
	sec = blk->speseg[j];
	
	while (sec != NULL) {
		for (k = 0; k < sec->chnum; k++) {
			if (sec->orient == '+') {
				tag1 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->beg,
											Spename[j], sec->chr, sec->orient, 
											"right", &b, &tmp);
				tag2 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->end,
											Spename[j], sec->chr, sec->orient,
											"left", &e, &tmp);
			}
			else {
				tag1 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->beg,
											Spename[j], sec->chr, sec->orient, 
											"right", &e, &tmp);
				tag2 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->end,
											Spename[j], sec->chr, sec->orient,
											"left", &b, &tmp);
			}
			if (tag1 != 0 && tag2 != 0)
				break;
		}
		if (k != sec->chnum)
			break;
		else 
			sec = sec->next;
	}
	if (sec == NULL || e <= b || tag1 == 0 || tag2 == 0) 
		invlist->speseg[j] = NULL;
	else {
		len = (int)(0.5*AAFEW*(ref->end - ref->beg));
		if (len != 0) {
				if (sec->orient == '+') {
					tag1 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->beg + len,
								Spename[j], sec->chr, sec->orient, 
								"right", &c, &tmp);
					tag2 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->end - len,
								Spename[j], sec->chr, sec->orient,
								"left", &f, &tmp);
				}
				else {
					tag1 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->beg + len,
								Spename[j], sec->chr, sec->orient, 
								"right", &f, &tmp);
					tag2 = mapbase(sec->cidlist[k], Spename[i], ref->chr, ref->end - len,
								Spename[j], sec->chr, sec->orient,
								"left", &c, &tmp);
				}
				if (f <= c || tag1 == 0 || tag2 == 0) {
					invlist->speseg[j] = NULL;
					return;
				}
		}
		newsg = (struct seg_list *)ckalloc(sizeof(struct seg_list));
		newsg->cidlist = NULL;
		newsg->next = NULL;
		strcpy(newsg->chr, sec->chr);
		newsg->orient = '+';
		newsg->beg = b;
		newsg->end = e;
		invlist->speseg[j] = newsg;
	}
}

int main(int argc, char* argv[]) {
		
	FILE *invfile, *tp;
	struct block_list *allblocks, *allinversion, *lst; 
	struct block_list *ilast, *pt, *newpt;
	struct seg_list  *ps;
	char buf[500], invlistfile[100];
	char rspe[50], rchr[50], sspe[50], schr[50];
	int rb, re, sb, se, left, right;
	int count, i, j, id, previd;

	if (argc != 3)
		fatal("arg = config-file blocks-list");

	get_spename(argv[1]);
	get_netdir(argv[1]);
	get_chaindir(argv[1]);
	rs = ref_spe_idx();
	
	allblocks = get_block_list(argv[2]);	
	
	allinversion = pt = ilast = NULL;
	
	for (i = 0; i < Spesz; i++) {
		if (i == rs)
			continue;
		sprintf(invlistfile, "__%s.%s.inp", Spename[rs], Spename[i]);
		fprintf(stderr, "opening %s\n", invlistfile);
		invfile = ckopen(invlistfile, "r");

		while(fgets(buf, 500, invfile)) {
			if (buf[0] == '\n')
				continue;
			if (buf[0] == '>') {
				if (sscanf(buf, ">%d", &id) != 1)
					fatalf("%s", buf);
				continue;
			}
			if (sscanf(buf, "%[^.].%[^:]:%d-%d %[^.].%[^:]:%d-%d %*c [%d %d]",
							rspe, rchr, &rb, &re, sspe, schr, &sb, &se, &left, &right) != 10)
				fatalf("%s", buf);
			if (left > rb || right < re) 
				fatalf("impossible: %d %d %d %d", left, rb, re, right);
			if (!same_string(rspe, Spename[rs]))
				fatalf("inconsistnet ref-spe %s vs %s", rspe, Spename[rs]);
			pt = get_inv_block(allinversion, rs, rchr, rb, re, left, right, &(lst));
			if (pt != NULL) {
				if (pt != lst)
					merge_blocks(pt, lst);
				ps = pt->speseg[rs];
				ps->beg = MIN(ps->beg, rb);
				ps->end = MAX(ps->end, re);
				add_info(pt, sspe, schr, sb, se, rs, left, right);
			}
			else if (Spetag[i] != 2){
				newpt = allocate_newblock();
				newpt->id = id;
				newpt->left = 0;
				newpt->right = BIGNUM;
				if (allinversion == NULL)
					allinversion = newpt;
				else {
					ilast = find_pos(allinversion, rs, rb, id);
					if (ilast == NULL) {
						newpt->next = allinversion;
						allinversion = newpt;
					}
					else {
						newpt->next = ilast->next;
						ilast->next = newpt;
					}
				}
				add_info(newpt, rspe, rchr, rb, re, rs, left, right);
				add_info(newpt, sspe, schr, sb, se, rs, left, right);
			}
		}
		fclose(invfile);
	}

	previd = count = 0;
	tp = ckopen("__tmp__file", "w");
	for (pt = allinversion; pt != NULL; pt = pt->next) {
		if (pt->id != previd)
			fprintf(tp, ">%d\n\n", pt->id);
	
		if (pt->right < pt->left)
			fprintf(tp, "XXX\n");
		
		fprintf(tp, "#%d %d %d\n", ++count, pt->left, pt->right);
		for (j = 0; j < Spesz; j++) {
			ps = pt->speseg[j];
			if (ps == NULL)
				fprintf(tp, " ? %*s\n", 13, Spename[j]);
			else {
				sprintf(buf, "%s.%s", Spename[j], ps->chr);
                fprintf(tp, " %c %*s:%d-%d\n", ps->orient, 13, buf, ps->beg, ps->end);
			}
		}
		fprintf(tp, "\n");
		previd = pt->id;
	}
	
	fclose(tp);
	
	for (i = 0; i < Spesz; i++) {
		for (pt = allinversion; pt != NULL; pt = pt->next) {
			if (i != rs && pt->speseg[i] == NULL)
				check_covered(pt, rs, i, allblocks);
		}
	}
	
	previd = count = 0;
	for (pt = allinversion; pt != NULL; pt = pt->next) {
		if (pt->id != previd)
			printf(">%d\n\n", pt->id);
		printf("#%d\n", ++count);
		for (j = 0; j < Spesz; j++) {
			ps = pt->speseg[j];
			if (ps == NULL)
				printf(" ? %*s\n", 13, Spename[j]);
			else {
				sprintf(buf, "%s.%s", Spename[j], ps->chr);
                printf(" %c %*s:%d-%d\n", ps->orient, 13, buf, ps->beg, ps->end);
			}
		}
		printf("\n");
		previd = pt->id;
	}
	
	free_block_list(allblocks);
	free_block_list(allinversion);
	
	return 0;
}
