#include "util.h"
#include "species.h"

#define MAXDEP	10
#define LE			0.5

struct net_list {
	char type[20], chrom[50];
	int level;
	int fbeg, fend;
	int sbeg, send;
	char orient;
	int qFar;
	struct net_list *next;
};

int get_level(char *s) {
	int i = 0;
	while (s[i] == ' ')
		++i; 
	return i;
}

void free_net_list(struct net_list *t) {
	struct net_list *p, *q;
	if (t != NULL) {
		p = t;
		for (;;) {
			q = p->next;
			p->next = NULL;
			free(p);
			if (q == NULL)
				break;
			else
				p = q;
		}
	}
}

struct net_list *get_net_list(char *fname) {
	FILE *fp;
	char buf[500], type[50], chrom[50], chr[50];
	struct net_list *head, *last, *p;
	char orient;
	int fbeg, fend, sbeg, send, qFar, level;
	
	head = last = NULL;

	fprintf(stderr, "opening %s\n", fname);
	fp = ckopen(fname, "r");
	
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#')
			continue;
		if (buf[0] == 'n' && buf[0] == 'e') {
			if (sscanf(buf, "net %s %*d", chr) != 1)
				fatalf("net-file %s: missing header", fname);
			continue; 
		}
		if (sscanf(buf, "%s %*s", type) != 1)
			fatalf("cannot parse net-file: %s", buf);
		if (same_string(type, "gap") 
			|| (same_string(type, "fill") && strstr(buf, "inv") != NULL)) {

			p = (struct net_list *)ckalloc(sizeof(struct net_list));
			p->next = NULL;
			if (head == NULL)
				head = last = p;
			else {
				last->next = p;
				last = p;
			}
			if (same_string(type, "gap"))
				strcpy(p->type, type);
			else 
				strcpy(p->type, "inv");
			level = get_level(buf);
			level /= 2;
			p->level = level;
		
			if (same_string(type, "gap")) {
				if (sscanf(buf, "%*s %d %d %s %c %d %d %*s",
							&fbeg, &fend, chrom, &orient, &sbeg, &send) != 6)
					fatalf("cannot parse net file: %s", buf);
				strcpy(p->chrom, chrom);
				p->fbeg = fbeg;
				p->fend = fbeg + fend;
				p->sbeg = sbeg;
				p->send = sbeg + send;
				p->orient = orient;
			}
			else if (same_string(type, "fill")) {
					if (sscanf(buf, 
"%*s %d %d %s %c %d %d id %*d score %*d ali %*d qOver %*d qFar %d qDup %*d type inv %*s",
						&fbeg, &fend, chrom, &orient, &sbeg, &send, &qFar) != 7)
						fatalf("cannot parse net-file: %s (inv)", buf);
					strcpy(p->chrom, chrom);
					p->fbeg = fbeg;
					p->fend = fbeg + fend;
					p->sbeg = sbeg;
					p->send = sbeg + send;
					p->qFar = qFar;
					p->orient = orient;
			}
		}
	}				
	fclose(fp);
	return head;
}

int check_inversion(struct block_list *blk, int idx, char *chrom,
													int beg, int end, int ref, int sec, char ori ) {
	int valid = 0;
	struct seg_list *p;
	
	for (p = blk->speseg[idx]; p != NULL; p = p->next)
		if (same_string(chrom, p->chr) && beg >= p->beg && end <= p->end)
			break;
	if (p != NULL){
		if (idx == ref)
			valid = 1;
		else if (idx == sec) {
			if (blk->speseg[ref]->orient == '+' && ori == ORT(blk->speseg[idx]->orient))
				valid = 1;
			if (blk->speseg[ref]->orient == '-' && ori == blk->speseg[idx]->orient)
				valid = 1;
		}
		else fatalf("impossible: %d %d %d", idx, ref, sec);
	}
	return valid;
}

int main(int argc, char* argv[]) {
	FILE *of;
	char netfilename[500], prevchrom[50], buf[500];
	char *refchrom;
	int rs, i, j, level;
	int fgbeg[MAXDEP], fgend[MAXDEP];
	struct block_list *blocklist, *blk;
	struct net_list *net, *nt;
	
	if (argc != 3)
		fatal("args = config-file block-list");
	
	get_spename(argv[1]);
	get_netdir(argv[1]);
	rs = ref_spe_idx();

	blocklist = get_block_list(argv[2]);

	net = NULL;
	
	for (i = 0; i < Spesz; i++) {
		if (i == rs)
			continue;
		prevchrom[0] = '\0';
		sprintf(buf, "__%s.%s.inp", Spename[rs], Spename[i]);
		of = ckopen(buf, "w");
		
		for (blk = blocklist; blk != NULL; blk = blk->next) {
			fprintf(stderr, "  processing block #%d\n", blk->id);
			fprintf(of, ">%d\n", blk->id);
		
			refchrom = (blk->speseg[rs])->chr;

			if (!same_string(prevchrom, refchrom)){
				if (net != NULL)
					free_net_list(net);
				sprintf(netfilename, "%s/%s/%s/net/%s.net", 
								Netdir, Spename[rs], Spename[i], refchrom);
				net = get_net_list(netfilename);
				strcpy(prevchrom, refchrom);
			}
		
			for (nt = net; nt != NULL; nt = nt->next) {
				if (same_string(nt->type, "gap")) {
					level = nt->level;
					fgbeg[level] = nt->fbeg;
					fgend[level] = nt->fend;
				}
				else if (same_string(nt->type, "inv")) {
					if (nt->fbeg > (blk->speseg[rs])->end)
						break;
					if (nt->qFar < (int)(LE * (nt->fend - nt->fbeg))) {
						if (check_inversion(blk, rs, refchrom, nt->fbeg, nt->fend, rs, i, nt->orient) 
				  	 		&& check_inversion(blk, i, nt->chrom, nt->sbeg, nt->send, rs, i, nt->orient)) {
							fprintf(of, "%s.%s:%d-%d %s.%s:%d-%d %c\t", 
										Spename[rs], refchrom, nt->fbeg, nt->fend, 
										Spename[i], nt->chrom, nt->sbeg, nt->send, nt->orient);
							j = nt->level;
							fprintf(of, "[%d %d]\n", fgbeg[j], fgend[j]);
						}
					}
				}
			}
			fprintf(of, "\n");
		}
		fclose(of);
	}
	free_net_list(net);
	free_block_list(blocklist);

	return 0;
}

